/* Core/apidefs.h - API constants */

#ifndef GOBI_CORE_APIDEFS_H
#define GOBI_CORE_APIDEFS_H

#ifdef __cplusplus
extern "C" {
#endif

typedef enum GobiType {
	GOBITYPE_UNKNOWN,
	GOBITYPE_2K,
	GOBITYPE_3K,
} GobiType;

#ifdef __cplusplus
};
#endif

#endif /* !GOBI_CORE_APIDEFS_H */

