/* Core/types.h - basic sdk types */

#ifndef GOBI_CORE_TYPES_H
#define GOBI_CORE_TYPES_H

typedef unsigned long      ULONG;
typedef unsigned long long ULONGLONG;
typedef signed char        INT8;
typedef unsigned char      BYTE;
typedef char               CHAR;
typedef unsigned short     WORD;
typedef unsigned short     USHORT;
typedef const char         *LPCSTR;

#endif /* !GOBI_CORE_TYPES_H */

